@@ -2,7 +2,7 @@
    producing program.bc
 
    Copyright (c) 2013 Kalle Raiskila
-                 2013-2017 Pekka Jääskeläinen
+                 2013-2019 Pekka Jääskeläinen
 
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
@@ -29,11 +29,15 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 
 #include "config.h"
 
-#include "clang/CodeGen/CodeGenAction.h"
-#include "clang/Frontend/CompilerInstance.h"
-#include "clang/Frontend/CompilerInvocation.h"
-#include "clang/Frontend/FrontendActions.h"
-#include "clang/Frontend/TextDiagnosticBuffer.h"
+#include <clang/Basic/Diagnostic.h>
+#include <clang/Driver/Compilation.h>
+#include <clang/Driver/Driver.h>
+#include <clang/CodeGen/CodeGenAction.h>
+#include <clang/Frontend/CompilerInstance.h>
+#include <clang/Frontend/CompilerInvocation.h>
+#include <clang/Frontend/FrontendActions.h>
+#include <clang/Frontend/TextDiagnosticBuffer.h>
+#include <clang/Frontend/TextDiagnosticPrinter.h>
 
 #ifndef LLVM_OLDER_THAN_4_0
 #include "clang/Lex/PreprocessorOptions.h"
@@ -69,7 +73,6 @@ using namespace llvm;
 
 POP_COMPILER_DIAGS
 
-
 /* Global pocl device to be used by passes if needed */
 cl_device_id currentPoclDevice = NULL;
 
@@ -109,8 +112,7 @@ load_source(FrontendOptions &fe,
 static inline int
 unlink_source(FrontendOptions &fe)
 {
-  // don't unlink in debug mode
-  if (pocl_get_bool_option("POCL_DEBUG", 0))
+  if (pocl_get_bool_option("POCL_LEAVE_KERNEL_COMPILER_TEMP_FILES", 0) != 0)
     return 0;
 
   FrontendInputFile const& file = fe.Inputs.front();
@@ -125,7 +127,7 @@ unlink_source(FrontendOptions &fe)
 static void appendToProgramBuildLog(cl_program program, unsigned device_i,
                                     std::string &s) {
   if (!s.empty()) {
-    POCL_MSG_ERR(s.c_str());
+    POCL_MSG_ERR("%s", s.c_str());
     /* this may not actually write anything if the buildhash is invalid,
      * but program->build_log still gets written.  */
     pocl_cache_append_to_buildlog(program, device_i, s.c_str(), s.size());
@@ -147,26 +149,28 @@ static void get_build_log(cl_program program,
                          unsigned device_i,
                          std::stringstream &ss_build_log,
                          clang::TextDiagnosticBuffer *diagsBuffer,
-                         const SourceManager &sm)
+                         const SourceManager *SM)
 {
-    for (TextDiagnosticBuffer::const_iterator i = diagsBuffer->err_begin(),
+  for (TextDiagnosticBuffer::const_iterator i = diagsBuffer->err_begin(),
          e = diagsBuffer->err_end(); i != e; ++i)
-      {
-        ss_build_log << "error: " << i->first.printToString(sm)
-                     << ": " << i->second << std::endl;
-      }
-    for (TextDiagnosticBuffer::const_iterator i = diagsBuffer->warn_begin(),
+  {
+    ss_build_log << "error: "
+                 << (SM == nullptr ? "" : (i->first.printToString(*SM) + ": "))
+                 << i->second << std::endl;
+  }
+  for (TextDiagnosticBuffer::const_iterator i = diagsBuffer->warn_begin(),
          e = diagsBuffer->warn_end(); i != e; ++i)
-      {
-        ss_build_log << "warning: " << i->first.printToString(sm)
-                     << ": " << i->second << std::endl;
-      }
+  {
+    ss_build_log << "warning: "
+                 << (SM == nullptr ? "" : (i->first.printToString(*SM) + ": "))
+                 << i->second << std::endl;
+  }
 
-    std::string log = ss_build_log.str();
-    appendToProgramBuildLog(program, device_i, log);
+  std::string log = ss_build_log.str();
+  appendToProgramBuildLog(program, device_i, log);
 }
 
-static llvm::Module *kernel_library(cl_device_id device);
+static llvm::Module *getKernelLibrary(cl_device_id device);
 
 int pocl_llvm_build_program(cl_program program,
                             unsigned device_i,
@@ -178,7 +182,6 @@ int pocl_llvm_build_program(cl_program program,
                             int linking_program)
 
 {
-  void* write_lock = NULL;
   char tempfile[POCL_FILENAME_LENGTH];
   tempfile[0] = 0;
   llvm::Module **mod = NULL;
@@ -278,15 +281,13 @@ int pocl_llvm_build_program(cl_program program,
 
   if (device->has_64bit_long)
     ss << "-Dcl_khr_int64 ";
-  // This can cause illegal optimizations when unaware
-  // of the barrier semantics. -O2 is the default opt level in
-  // Clang for OpenCL C and seems to affect the performance
-  // of the end result, even if we optimize the final WG
-  // func. TODO: There should be 'noduplicate' etc. flags in
-  // the 'barrier' function to prevent them.
-  // ss << "-O2 ";
 
-  ss << "-x cl ";
+  ss << "-DPOCL_DEVICE_ADDRESS_BITS=" << device->address_bits << " ";
+#ifndef LLVM_OLDER_THAN_4_0
+  ss << "-D__USE_CLANG_OPENCL_C_H ";
+#endif
+
+  ss << "-xcl ";
   // Remove the inline keywords to force the user functions
   // to be included in the program. Otherwise they will
   // be removed and not inlined due to -O0.
@@ -353,9 +354,9 @@ int pocl_llvm_build_program(cl_program program,
   }
 
   for (unsigned idx = 0; idx < itemstrs.size(); idx++) {
-      // note: if itemstrs is modified after this, itemcstrs will be full
-      // of invalid pointers! Could make copies, but would have to clean up then...
-      itemcstrs.push_back(itemstrs[idx].c_str());
+    // note: if itemstrs is modified after this, itemcstrs will be full
+    // of invalid pointers! Could make copies, but would have to clean up then...
+    itemcstrs.push_back(itemstrs[idx].c_str());
   }
 
 #ifdef DEBUG_POCL_LLVM_API
@@ -373,7 +374,8 @@ int pocl_llvm_build_program(cl_program program,
        diags)) {
     pocl_cache_create_program_cachedir(program, device_i, NULL, 0,
                                        program_bc_path);
-    get_build_log(program, device_i, ss_build_log, diagsBuffer, CI.getSourceManager());
+    get_build_log(program, device_i, ss_build_log, diagsBuffer,
+                  CI.hasSourceManager() ? &CI.getSourceManager() : nullptr);
     return CL_INVALID_BUILD_OPTIONS;
   }
 
@@ -408,36 +410,37 @@ int pocl_llvm_build_program(cl_program program,
   la->MathErrno = false; // -fno-math-errno
   la->NoBuiltin = true;  // -fno-builtin
   la->AsmBlocks = true;  // -fasm (?)
-  la->setStackProtector(LangOptions::StackProtectorMode::SSPOff);
-  la->PICLevel = 0;
-#ifdef LLVM_OLDER_THAN_3_9
-  la->PIELevel = 0;
-#else
-  la->PIE = 0;
-#endif
 
-  std::string kernelh;
+  la->setStackProtector(LangOptions::StackProtectorMode::SSPOff);
+
+  la->PICLevel = PICLevel::BigPIC;
+  la->PIE = 1;
+
+  std::string IncludeRoot;
+  std::string KernelH;
   std::string BuiltinRenamesH;
+  std::string PoclTypesH;
 
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
-    kernelh  = SRCDIR;
+    IncludeRoot = SRCDIR;
 #else
   if (0) {
 #endif
   } else {
-    kernelh = POCL_INSTALL_PRIVATE_DATADIR;
+    IncludeRoot = POCL_INSTALL_PRIVATE_DATADIR;
   }
-  BuiltinRenamesH = kernelh;
-  kernelh += "/include/_kernel.h";
-  BuiltinRenamesH += "/include/_builtin_renames.h";
+  KernelH = IncludeRoot + "/include/_kernel.h";
+  BuiltinRenamesH = IncludeRoot + "/include/_builtin_renames.h";
+  PoclTypesH = IncludeRoot + "/include/pocl_types.h";
 
+  po.Includes.push_back(PoclTypesH);
   po.Includes.push_back(BuiltinRenamesH);
 #ifndef LLVM_OLDER_THAN_4_0
   // Use Clang's opencl-c.h header.
   po.Includes.push_back(CLANG_RESOURCE_DIR "/include/opencl-c.h");
 #endif
-  po.Includes.push_back(kernelh);
+  po.Includes.push_back(KernelH);
   clang::TargetOptions &ta = pocl_build.getTargetOpts();
   ta.Triple = device->llvm_target_triplet;
   if (device->llvm_cpu != NULL)
@@ -466,7 +469,11 @@ int pocl_llvm_build_program(cl_program program,
   // This workarounds a Frontend codegen issues with an illegal address
   // space cast which is later flattened (and thus implicitly fixed) in
   // the TargetAddressSpaces. See:  https://github.com/pocl/pocl/issues/195
+#ifdef POCL_USE_FAKE_ADDR_SPACE_IDS
   cg.VerifyModule = false;
+#else
+  cg.VerifyModule = true;
+#endif
 
   PreprocessorOutputOptions &poo = pocl_build.getPreprocessorOutputOpts();
   poo.ShowCPP = 1;
@@ -476,7 +483,7 @@ int pocl_llvm_build_program(cl_program program,
   poo.ShowMacros = 1;
   poo.RewriteIncludes = 0;
 
-  pocl_cache_tempname(tempfile, ".cl", NULL);
+  pocl_cache_tempname(tempfile, ".preproc.cl", NULL);
   fe.OutputFile.assign(tempfile);
 
   bool success = true;
@@ -488,16 +495,19 @@ int pocl_llvm_build_program(cl_program program,
   if (success) {
     pocl_read_file(tempfile, &PreprocessedOut, &PreprocessedSize);
   }
+  /* always remove preprocessed output - the sources are in different files */
+  pocl_remove(tempfile);
+
   if (pocl_get_bool_option("POCL_LEAVE_KERNEL_COMPILER_TEMP_FILES", 0) == 0) {
     if (num_input_headers > 0)
       pocl_rm_rf(temp_include_dir);
-    pocl_remove(tempfile);
   }
 
   if (PreprocessedOut == nullptr) {
     pocl_cache_create_program_cachedir(program, device_i, NULL, 0,
                                        program_bc_path);
-    get_build_log(program, device_i, ss_build_log, diagsBuffer, CI.getSourceManager());
+    get_build_log(program, device_i, ss_build_log, diagsBuffer,
+                  CI.hasSourceManager() ? &CI.getSourceManager() : nullptr);
     return CL_BUILD_PROGRAM_FAILURE;
   }
 
@@ -506,10 +516,10 @@ int pocl_llvm_build_program(cl_program program,
 
   POCL_MEM_FREE(PreprocessedOut);
 
-  if (pocl_exists(program_bc_path)) {
-    unlink_source(fe);
+  unlink_source(fe);
+
+  if (pocl_exists(program_bc_path))
     return CL_SUCCESS;
-  }
 
   // TODO: use pch: it is possible to disable the strict checking for
   // the compilation flags used to compile it and the current translation
@@ -518,9 +528,7 @@ int pocl_llvm_build_program(cl_program program,
   clang::EmitLLVMOnlyAction EmitLLVM(&c);
   success = CI.ExecuteAction(EmitLLVM);
 
-  unlink_source(fe);
-
-  get_build_log(program, device_i, ss_build_log, diagsBuffer, CI.getSourceManager());
+  get_build_log(program, device_i, ss_build_log, diagsBuffer, &CI.getSourceManager());
 
   if (!success)
     return CL_BUILD_PROGRAM_FAILURE;
@@ -538,24 +546,17 @@ int pocl_llvm_build_program(cl_program program,
 
   ++numberOfIRs;
 
-  // a workaround for errors with PIC + variables in constant addrspace
-  // fails on the test_convert_type_X tests with this error:
-  //   relocation R_X86_64_PC32 against symbol `char_values' can not be used
-  //   when making a shared object; recompile with -fPIC
-#ifdef LLVM_OLDER_THAN_3_9
-  (*mod)->setPICLevel(PICLevel::Default);
-#else
-  (*mod)->setPICLevel(PICLevel::NotPIC);
-  (*mod)->setPIELevel(PIELevel::Default);
-#endif
-
+  if ((*mod)->getModuleFlag("PIC Level") == nullptr)
+    (*mod)->setPICLevel(PICLevel::BigPIC);
+  if ((*mod)->getModuleFlag("PIE Level") == nullptr)
+    (*mod)->setPIELevel(PIELevel::Large);
 
   // link w kernel lib, but not if we're called from clCompileProgram()
   // Later this should be replaced with indexed linking of source code
   // and/or bitcode for each kernel.
   if (linking_program) {
     currentPoclDevice = device;
-    llvm::Module *libmodule = kernel_library(device);
+    llvm::Module *libmodule = getKernelLibrary(device);
     assert(libmodule != NULL);
     std::string log("Error(s) while linking: \n");
     if (link(*mod, libmodule, log)) {
@@ -569,9 +570,6 @@ int pocl_llvm_build_program(cl_program program,
     }
   }
 
-  write_lock = pocl_cache_acquire_writer_lock_i(program, device_i);
-  assert(write_lock);
-
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
   /* Always retain program.bc. Its required in clBuildProgram */
@@ -593,31 +591,67 @@ int pocl_llvm_build_program(cl_program program,
   program->binaries[device_i] = (unsigned char *) malloc(n);
   std::memcpy(program->binaries[device_i], content.c_str(), n);
 
-  pocl_cache_release_lock(write_lock);
-
   return CL_SUCCESS;
 }
 
-
-int pocl_llvm_link_program(cl_program program,
-                           unsigned device_i,
-                           char *program_bc_path,
-                           cl_uint num_input_programs,
+int pocl_llvm_link_program(cl_program program, unsigned device_i,
+                           char *program_bc_path, cl_uint num_input_programs,
                            unsigned char **cur_device_binaries,
-                           size_t *cur_device_binary_sizes,
-                           void **cur_llvm_irs,
-                           int create_library) {
+                           size_t *cur_device_binary_sizes, void **cur_llvm_irs,
+                           int create_library, int spir) {
 
-  void *write_lock;
   std::string concated_binaries;
+  llvm::Module *linked_module = nullptr;
   size_t n = 0, i;
   cl_device_id device = program->devices[device_i];
   llvm::Module **modptr = (llvm::Module **)&program->llvm_irs[device_i];
   int error;
 
+  currentPoclDevice = device;
+  llvm::Module *libmodule = getKernelLibrary(device);
+  assert(libmodule != NULL);
+
   PoclCompilerMutexGuard lockHolder(NULL);
   InitializeLLVM();
 
+  if (spir) {
+#ifdef ENABLE_SPIR
+
+    POCL_RETURN_ERROR_ON ((device->endian_little == CL_FALSE),
+                         CL_LINK_PROGRAM_FAILURE,
+                         "SPIR is only supported on little-endian devices\n");
+
+    linked_module = parseModuleIRMem((char *)program->binaries[device_i],
+                                     program->binary_sizes[device_i]);
+
+    const std::string &spir_triple = linked_module->getTargetTriple();
+    size_t spir_addrbits = Triple(spir_triple).isArch64Bit() ? 64 : 32;
+
+    if (device->address_bits != spir_addrbits) {
+        delete linked_module;
+        POCL_RETURN_ERROR_ON (1, CL_LINK_PROGRAM_FAILURE,
+                       "Device address bits != SPIR binary triple address "
+                       "bits, device: %s / module: %s\n",
+                       device->llvm_target_triplet, spir_triple.c_str());
+    }
+
+    /* Note this is a hack to get SPIR working. We'll be linking the
+     * host kernel library (plain LLVM IR) to the SPIR program.bc,
+     * so LLVM complains about incompatible DataLayouts.
+     */
+    linked_module->setTargetTriple(libmodule->getTargetTriple());
+    linked_module->setDataLayout(libmodule->getDataLayout());
+    if (linked_module->getModuleFlag("PIC Level") == nullptr)
+      linked_module->setPICLevel(PICLevel::BigPIC);
+    if (linked_module->getModuleFlag("PIE Level") == nullptr)
+      linked_module->setPIELevel(PIELevel::Large);
+
+#else
+    POCL_MSG_ERR("SPIR not supported\n");
+    return CL_LINK_PROGRAM_FAILURE;
+#endif
+  } else {
+
 #ifdef LLVM_OLDER_THAN_3_8
   llvm::Module *mod =
       new llvm::Module(StringRef("linked_program"), GlobalContext());
@@ -638,8 +672,10 @@ int pocl_llvm_link_program(cl_program program,
 #ifdef LLVM_OLDER_THAN_3_8
     if (Linker::LinkModules(mod, llvm::CloneModule(p))) {
       delete mod;
-#else
+#elif LLVM_OLDER_THAN_7_0
     if (Linker::linkModules(*mod, llvm::CloneModule(p))) {
+#else
+    if (Linker::linkModules(*mod, llvm::CloneModule(*p))) {
 #endif
       std::string msg = getDiagString();
       appendToProgramBuildLog(program, device_i, msg);
@@ -648,10 +684,11 @@ int pocl_llvm_link_program(cl_program program,
   }
 
 #ifdef LLVM_OLDER_THAN_3_8
-  llvm::Module *linked_module = mod;
+  linked_module = mod;
 #else
-  llvm::Module *linked_module = mod.release();
+  linked_module = mod.release();
 #endif
+  }
 
   if (linked_module == nullptr)
     return CL_LINK_PROGRAM_FAILURE;
@@ -664,9 +701,6 @@ int pocl_llvm_link_program(cl_program program,
 
   if (!create_library) {
     // linked all the programs together, now link in the kernel library
-    currentPoclDevice = device;
-    llvm::Module *libmodule = kernel_library(device);
-    assert(libmodule != NULL);
     std::string log("Error(s) while linking: \n");
     if (link(linked_module, libmodule, log)) {
       appendToProgramBuildLog(program, device_i, log);
@@ -693,9 +727,6 @@ int pocl_llvm_link_program(cl_program program,
       return error;
     }
 
-  write_lock = pocl_cache_acquire_writer_lock_i(program, device_i);
-  assert(write_lock);
-
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
   /* Always retain program.bc. Its required in clBuildProgram */
@@ -717,8 +748,6 @@ int pocl_llvm_link_program(cl_program program,
   program->binaries[device_i] = (unsigned char *)malloc(n);
   std::memcpy(program->binaries[device_i], content.c_str(), n);
 
-  pocl_cache_release_lock(write_lock);
-
   return CL_SUCCESS;
 }
 
@@ -732,7 +761,7 @@ const char *getX86KernelLibName() {
   if (Features["sse2"])
     res = "sse2";
   else
-    POCL_ABORT("Pocl on x86_64 requires at least SSE2");
+    POCL_ABORT("Pocl on x86_64 requires at least SSE2\n");
   if (Features["ssse3"] && Features["cx16"])
     res = "ssse3";
   if (Features["sse4.1"] && Features["cx16"])
@@ -762,9 +791,7 @@ static std::map<cl_device_id, llvm::Module *> kernelLibraryMap;
  * Return the OpenCL C built-in function library bitcode
  * for the given device.
  */
-static llvm::Module*
-kernel_library
-(cl_device_id device)
+static llvm::Module* getKernelLibrary(cl_device_id device)
 {
   Triple triple(device->llvm_target_triplet);
 
@@ -773,8 +800,10 @@ kernel_library
 
   const char *subdir = "host";
   bool is_host = true;
+  // TODO: move this to the device layer, a property to ask for
+  // the kernel builtin bitcode library name, including its subdir
 #ifdef TCE_AVAILABLE
-  if (triple.getArch() == Triple::tce) {
+  if (triple.getArch() == Triple::tce || triple.getArch() == Triple::tcele) {
     subdir = "tce";
     is_host = false;
   }
@@ -799,55 +828,32 @@ kernel_library
   }
 #endif
 
-  // TODO sync with Nat Ferrus' indexed linking
   std::string kernellib;
   std::string kernellib_fallback;
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
     kernellib = BUILDDIR;
-#else
-  if (0) {
-#endif
     kernellib += "/lib/kernel/";
     kernellib += subdir;
-    // TODO: get this from the TCE target triplet
-    kernellib += "/kernel-";
-    kernellib += device->llvm_target_triplet;
-    if (is_host) {
-      kernellib += '-';
-      kernellib_fallback = kernellib;
-      kernellib_fallback += OCL_KERNEL_TARGET_CPU;
-      kernellib_fallback += ".bc";
-#ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-      if (triple.getArch() == Triple::x86_64 ||
-          triple.getArch() == Triple::x86)
-        kernellib += getX86KernelLibName();
-      else
+  } else // POCL_BUILDING == 0, use install dir
 #endif
-        kernellib += device->llvm_cpu;
-    }
-  } else { // POCL_BUILDING == 0, use install dir
-    kernellib = POCL_INSTALL_PRIVATE_DATADIR;
-    kernellib += "/kernel-";
-    kernellib += device->llvm_target_triplet;
-    if (is_host) {
-      kernellib += '-';
-      kernellib_fallback = kernellib;
-      kernellib_fallback += OCL_KERNEL_TARGET_CPU;
-      kernellib_fallback += ".bc";
+  kernellib = POCL_INSTALL_PRIVATE_DATADIR;
+  kernellib += "/kernel-";
+  kernellib += device->llvm_target_triplet;
+  if (is_host) {
+    kernellib += '-';
 #ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-      if (triple.getArch() == Triple::x86_64 ||
-          triple.getArch() == Triple::x86)
-        kernellib += getX86KernelLibName();
-      else
+    kernellib += getX86KernelLibName();
+#else
+    kernellib_fallback = kernellib;
+    kernellib_fallback += OCL_KERNEL_TARGET_CPU;
+    kernellib_fallback += ".bc";
+    kernellib += device->llvm_cpu;
 #endif
-        kernellib += device->llvm_cpu;
-    }
   }
   kernellib += ".bc";
 
   llvm::Module *lib;
-  SMDiagnostic Err;
 
   if (pocl_exists(kernellib.c_str()))
     {
@@ -856,6 +862,7 @@ kernel_library
     }
   else
     {
+#ifndef KERNELLIB_HOST_DISTRO_VARIANTS
       if (is_host && pocl_exists(kernellib_fallback.c_str()))
         {
           POCL_MSG_WARN("Using fallback %s as the built-in lib.\n",
@@ -863,7 +870,8 @@ kernel_library
           lib = parseModuleIR(kernellib_fallback.c_str());
         }
       else
-        POCL_ABORT("Kernel library file %s doesn't exist.", kernellib.c_str());
+#endif
+        POCL_ABORT("Kernel library file %s doesn't exist.\n", kernellib.c_str());
     }
   assert (lib != NULL);
   kernelLibraryMap[device] = lib;
@@ -878,3 +886,47 @@ void cleanKernelLibrary() {
   }
   kernelLibraryMap.clear();
 }
+
+#ifndef LLVM_OLDER_THAN_5_0
+/**
+ * Invoke the Clang compiler through its Driver API.
+ *
+ * @param Device the device of which toolchain to use.
+ * @param Args the command line arguments that would be passed to Clang
+ *             (a NULL terminated list). Args[0] should be the path to
+ *             the Clang binary.
+ * @return 0 on success, error code otherwise.
+ */
+int pocl_invoke_clang(cl_device_id Device, const char** Args) {
+
+  // Borrowed from driver.cpp (clang driver). We do not really care about
+  // diagnostics, but just want to get the compilation command invoked with
+  // the target's toolchain as defined in Clang.
+  IntrusiveRefCntPtr<DiagnosticOptions> DiagOpts = new DiagnosticOptions;
+
+  TextDiagnosticPrinter *DiagClient =
+    new TextDiagnosticPrinter(llvm::errs(), &*DiagOpts);
+
+  IntrusiveRefCntPtr<DiagnosticIDs> DiagID(new DiagnosticIDs());
+
+  DiagnosticsEngine Diags(DiagID, &*DiagOpts, DiagClient);
+
+  driver::Driver TheDriver(CLANG, Device->llvm_target_triplet, Diags);
+
+  const char **ArgsEnd = Args;
+  while (*ArgsEnd++ != nullptr) {}
+
+  llvm::ArrayRef<const char*> ArgsArray(Args, ArgsEnd);
+
+  std::unique_ptr<driver::Compilation> C(
+    TheDriver.BuildCompilation(ArgsArray));
+
+  if (C && !C->containsError()) {
+    SmallVector<std::pair<int, const driver::Command *>, 4> FailingCommands;
+    return TheDriver.ExecuteCompilation(*C, FailingCommands);
+  } else {
+    return -1;
+  }
+
+}
+#endif
