@@ -20,12 +20,12 @@
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
 */
+#include "poclu.h"
 #include <CL/cl.h>
-#include <string.h>
-#include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include "poclu.h"
+#include <string.h>
+
 #include "config.h"
 #include "pocl.h"
 
@@ -49,10 +49,17 @@ char kernelSourceCode[] =
 
 #define BUF_LEN 2000
 
-#define SPIR_FILE(NUM, SUFFIX) SPIR_FILE_2(NUM) SUFFIX
-#define SPIR_FILE_2(NUM) SRCDIR "/tests/runtime/clGetKernelArgInfo.spir" #NUM
+static const char *nometa_files[]
+    = { SRCDIR "/tests/runtime/clGetKernelArgInfo.spir32_nometa",
+        SRCDIR "/tests/runtime/clGetKernelArgInfo.spir64_nometa" };
 
-int test_program(cl_program program, int is_spir) {
+static const char *meta_files[]
+    = { SRCDIR "/tests/runtime/clGetKernelArgInfo.spir32_meta",
+        SRCDIR "/tests/runtime/clGetKernelArgInfo.spir64_meta" };
+
+int
+test_program (cl_program program, int is_spir)
+{
 
   cl_int err;
   size_t retsize;
@@ -125,31 +132,34 @@ int test_program(cl_program program, int is_spir) {
   err = clGetKernelArgInfo(test_kernel, 0, CL_KERNEL_ARG_TYPE_NAME,
                             BUF_LEN, &kernel_arg.string, &retsize);
   CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-  TEST_ASSERT((retsize==6) && " arg type name size of test_kernel doesnt fit");
-  TEST_ASSERT((strncmp(kernel_arg.string, "char*", 6)==0) &&
-	      " arg type name of test_kernel doesnt compare");
-
+  TEST_ASSERT ((retsize == 6)
+               && " arg type name size of test_kernel doesn't fit");
+  TEST_ASSERT ((strncmp (kernel_arg.string, "char*", 6) == 0)
+               && " arg type name of test_kernel doesn't compare");
 
   err = clGetKernelArgInfo(test_kernel, 1, CL_KERNEL_ARG_TYPE_NAME,
                             BUF_LEN, &kernel_arg.string, &retsize);
   CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-  TEST_ASSERT((retsize==7) && " arg type name size of test_kernel doesnt fit");
-  TEST_ASSERT((strncmp(kernel_arg.string, "float*", 7)==0) &&
-	      " arg type name of test_kernel doesnt compare");
+  TEST_ASSERT ((retsize == 7)
+               && " arg type name size of test_kernel doesn't fit");
+  TEST_ASSERT ((strncmp (kernel_arg.string, "float*", 7) == 0)
+               && " arg type name of test_kernel doesn't compare");
 
   err = clGetKernelArgInfo(test_kernel, 3, CL_KERNEL_ARG_TYPE_NAME,
                             BUF_LEN, &kernel_arg.string, &retsize);
   CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-  TEST_ASSERT((retsize==6) && " arg type name size of test_kernel doesnt fit");
-  TEST_ASSERT((strncmp(kernel_arg.string, "float", 6)==0) &&
-	      " arg type name of test_kernel doesnt compare");
+  TEST_ASSERT ((retsize == 6)
+               && " arg type name size of test_kernel doesn't fit");
+  TEST_ASSERT ((strncmp (kernel_arg.string, "float", 6) == 0)
+               && " arg type name of test_kernel doesn't compare");
 
   err = clGetKernelArgInfo(test_kernel, 4, CL_KERNEL_ARG_TYPE_NAME,
                             BUF_LEN, &kernel_arg.string, &retsize);
   CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-  TEST_ASSERT((retsize==5) && " arg type name size of test_kernel doesnt fit");
-  TEST_ASSERT((strncmp(kernel_arg.string, "int*", 5)==0) &&
-	      " arg type name of test_kernel doesnt compare");
+  TEST_ASSERT ((retsize == 5)
+               && " arg type name size of test_kernel doesn't fit");
+  TEST_ASSERT ((strncmp (kernel_arg.string, "int*", 5) == 0)
+               && " arg type name of test_kernel doesn't compare");
 
   /* TYPE QUALIFIER tests */
   // constant char* msg, global volatile float* in, global float* out, const float j, local int* c
@@ -218,8 +228,10 @@ int test_program(cl_program program, int is_spir) {
     printf("arg name not available (this is normal for SPIR)\n");
   } else {
     CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-    TEST_ASSERT((retsize==4) && " arg name size of test_kernel doesnt fit");
-    TEST_ASSERT((strncmp(kernel_arg.string, "msg", 4)==0) && " arg name of test_kernel doesnt compare");
+    TEST_ASSERT ((retsize == 4)
+                 && " arg name size of test_kernel doesn't fit");
+    TEST_ASSERT ((strncmp (kernel_arg.string, "msg", 4) == 0)
+                 && " arg name of test_kernel doesn't compare");
   }
 
   err = clGetKernelArgInfo(test_kernel, 1, CL_KERNEL_ARG_NAME,
@@ -228,8 +240,10 @@ int test_program(cl_program program, int is_spir) {
     printf("arg name not available (this is normal for SPIR)\n");
   } else {
     CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-    TEST_ASSERT((retsize==3) && " arg name size of test_kernel doesnt fit");
-    TEST_ASSERT((strncmp(kernel_arg.string, "in", 3)==0) && " arg name of test_kernel doesnt compare");
+    TEST_ASSERT ((retsize == 3)
+                 && " arg name size of test_kernel doesn't fit");
+    TEST_ASSERT ((strncmp (kernel_arg.string, "in", 3) == 0)
+                 && " arg name of test_kernel doesn't compare");
   }
 
   err = clGetKernelArgInfo(test_kernel, 2, CL_KERNEL_ARG_NAME,
@@ -238,8 +252,10 @@ int test_program(cl_program program, int is_spir) {
     printf("arg name not available (this is normal for SPIR)\n");
   } else {
     CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-    TEST_ASSERT((retsize==4) && " arg name size of test_kernel doesnt fit");
-    TEST_ASSERT((strncmp(kernel_arg.string, "out", 4)==0) && " arg name of test_kernel doesnt compare");
+    TEST_ASSERT ((retsize == 4)
+                 && " arg name size of test_kernel doesn't fit");
+    TEST_ASSERT ((strncmp (kernel_arg.string, "out", 4) == 0)
+                 && " arg name of test_kernel doesn't compare");
   }
 
   err = clGetKernelArgInfo(test_kernel, 3, CL_KERNEL_ARG_NAME,
@@ -248,8 +264,10 @@ int test_program(cl_program program, int is_spir) {
     printf("arg name not available (this is normal for SPIR)\n");
   } else {
     CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-    TEST_ASSERT((retsize==2) && " arg name size of test_kernel doesnt fit");
-    TEST_ASSERT((strncmp(kernel_arg.string, "j", 2)==0) && " arg name of test_kernel doesnt compare");
+    TEST_ASSERT ((retsize == 2)
+                 && " arg name size of test_kernel doesn't fit");
+    TEST_ASSERT ((strncmp (kernel_arg.string, "j", 2) == 0)
+                 && " arg name of test_kernel doesn't compare");
   }
 
   /* ACCESS QUALIFIER tests for test_kernel2 */
@@ -274,17 +292,19 @@ int test_program(cl_program program, int is_spir) {
   err = clGetKernelArgInfo(test_kernel2, 0, CL_KERNEL_ARG_TYPE_NAME,
                             BUF_LEN, &kernel_arg.string, &retsize);
   CHECK_OPENCL_ERROR_IN("clGetKernelArgInfo");
-  TEST_ASSERT((retsize==12) && " arg type name size of test_kernel doesnt fit");
-  TEST_ASSERT((strncmp(kernel_arg.string, "tasty_float", 12)==0) && " arg type name of test_kernel2 doesnt compare");
+  TEST_ASSERT ((retsize == 12)
+               && " arg type name size of test_kernel doesn't fit");
+  TEST_ASSERT ((strncmp (kernel_arg.string, "tasty_float", 12) == 0)
+               && " arg type name of test_kernel2 doesn't compare");
 
   err = clReleaseKernel(test_kernel);
   CHECK_OPENCL_ERROR_IN("clReleaseKernel");
   err = clReleaseKernel(test_kernel2);
   CHECK_OPENCL_ERROR_IN("clReleaseKernel");
   return EXIT_SUCCESS;
-
 }
 
+/*
 int test_program_nometa(cl_program program) {
 
   cl_int err;
@@ -301,8 +321,6 @@ int test_program_nometa(cl_program program) {
   CHECK_OPENCL_ERROR_IN("clCreateKernel");
 
 
-  /* ADDR SPACE QUALIFIER tests */
-  // constant char* msg, global volatile float* in, global float* out, const float j, local int* c
   err = clGetKernelArgInfo(test_kernel, 0, CL_KERNEL_ARG_ADDRESS_QUALIFIER,
                             BUF_LEN, &kernel_arg.address, &retsize);
   TEST_ASSERT(err == CL_KERNEL_ARG_INFO_NOT_AVAILABLE);
@@ -324,8 +342,12 @@ int test_program_nometa(cl_program program) {
 
   return EXIT_SUCCESS;
 }
+*/
 
-int spir_program(char * filename, cl_context ctx, cl_device_id did, cl_program* program) {
+int
+spir_program (const char *filename, cl_context ctx, cl_device_id did,
+              cl_program *program)
+{
   cl_int err;
 
   size_t program_size;
@@ -372,8 +394,12 @@ int main()
   program_size = strlen (kernelSourceCode);
   program_buffer = kernelSourceCode;
 
-  program = clCreateProgramWithSource (ctx, 1,
-                                       (const char**)&program_buffer,
+  cl_uint address_bits;
+  err = clGetDeviceInfo (did, CL_DEVICE_ADDRESS_BITS, sizeof (address_bits),
+                         &address_bits, NULL);
+  CHECK_OPENCL_ERROR_IN ("clGetDeviceInfo");
+
+  program = clCreateProgramWithSource (ctx, 1, (const char **)&program_buffer,
                                        &program_size, &err);
   CHECK_OPENCL_ERROR_IN("clCreateProgramWithBinary");
   TEST_ASSERT(program);
@@ -397,19 +423,26 @@ int main()
 
   /* SPIR program */
 
-  printf("\nSPIR with metadata\n");
-  TEST_ASSERT(spir_program(SPIR_FILE(POCL_DEVICE_ADDRESS_BITS, "_meta"), ctx, did, &program) == EXIT_SUCCESS);
+  printf ("\nSPIR with metadata\n");
 
-  TEST_ASSERT(test_program(program, 1) == EXIT_SUCCESS);
+  const char *filename = (address_bits == 32 ? meta_files[0] : meta_files[1]);
+  TEST_ASSERT (spir_program (filename, ctx, did, &program) == EXIT_SUCCESS);
+
+  TEST_ASSERT (test_program (program, 1) == EXIT_SUCCESS);
 
   CHECK_CL_ERROR(clReleaseProgram(program));
 
-  printf("\nSPIR WITHOUT metadata\n");
-  TEST_ASSERT(spir_program(SPIR_FILE(POCL_DEVICE_ADDRESS_BITS, "_nometa"), ctx, did, &program) == EXIT_SUCCESS);
+  /* SPIR program without metadata - currently disabled since Clang seems to
+   * always generate metadata. */
+  /*
+    printf("\nSPIR WITHOUT metadata\n");
+    filename = (address_bits == 32 ? nometa_files[0] : nometa_files[1]);
+    TEST_ASSERT(spir_program(filename, ctx, did, &program) == EXIT_SUCCESS);
 
-  TEST_ASSERT(test_program_nometa(program) == EXIT_SUCCESS);
+    TEST_ASSERT(test_program_nometa(program) == EXIT_SUCCESS);
 
-  CHECK_CL_ERROR(clReleaseProgram(program));
+    CHECK_CL_ERROR(clReleaseProgram(program));
+  */
 
 FINISH:
   CHECK_CL_ERROR (clReleaseCommandQueue (queue));
